/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.basis.solver.integer;

import java.math.BigInteger;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * {@link IntegerEquation} represents an equality or inequality of two integer variables
 * (objects of {@link IntegerVariable}).
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class IntegerEquation<V> {
  /** The equality/inequality flag. */
  public boolean equal;
  /** The variable-value/variable-variable flag. */
  public boolean value;
  /** The left-hand-side variable. */
  public V lhs;
  /** The right-hand-side variable. */
  public V rhs;
  /** The right-hand-side value. */
  public BigInteger val;

  /**
   * Constructs an equality/inequality.
   * 
   * @param lhs the left-hand-side variable.
   * @param rhs the right-hand-side variable.
   * @param equal the equality/inequality flag.
   * @throws IllegalArgumentException if {@code lhs} or {@code rhs} is null.
   */
  public IntegerEquation(
      final V lhs, final V rhs, final boolean equal) {
    InvariantChecks.checkNotNull(lhs);
    InvariantChecks.checkNotNull(rhs);

    this.lhs = lhs;
    this.rhs = rhs;
    this.equal = equal;
    this.value = false;
  }

  /**
   * Constructs an equality/inequality.
   * 
   * @param lhs the left-hand-side variable.
   * @param rhs the right-hand-side value.
   * @param equal the equality/inequality flag.
   * @throws IllegalArgumentException if {@code lhs} or {@code rhs} is null.
   */
  public IntegerEquation(
      final V lhs, final BigInteger rhs, final boolean equal) {
    InvariantChecks.checkNotNull(lhs);
    InvariantChecks.checkNotNull(rhs);

    this.lhs = lhs;
    this.val = rhs;
    this.equal = equal;
    this.value = true;
  }

  /**
   * Checks whether this equation contradicts to the given one.
   * 
   * @param equation the equation to be matched with this one.
   * @return {@code true} if this equation definitely contradicts to the given one;
   *         {@code false} if this equation seems to be consistent to the given one. 
   */
  public boolean contradictsTo(final IntegerEquation<V> equation) {
    InvariantChecks.checkNotNull(equation);

    if (!lhs.equals(equation.lhs) && !lhs.equals(equation.rhs)) {
      return false;
    }

    if (value && equation.value) {
      if ((val.compareTo(equation.val) == 0) && (equal != equation.equal)) {
        return true;
      }
      if ((val.compareTo(equation.val) != 0) && equal && equation.equal) {
        return true;
      }
    }

    if (!value && !equation.value) {
      final V equationRhs = lhs.equals(equation.lhs) ? equation.rhs : equation.lhs;

      if (rhs.equals(equationRhs) && equal != equation.equal) {
        return true;
      }
    }

    return false;
  }

  /**
   * Checks whether this equation ({@code A}) is stronger than the given one ({@code B}), i.e.
   * the property {@code A => B} holds.
   * 
   * @param equation the equation to be matched with this one.
   * @return {@code true} if this equation is definitely stronger than the given one;
   *         {@code false} if this equation does not seem to be stronger than the given one. 
   */
  public boolean strongerThan(final IntegerEquation<V> equation) {
    InvariantChecks.checkNotNull(equation);

    if (!lhs.equals(equation.lhs) && !lhs.equals(equation.rhs)) {
      return false;
    }

    if (value && equation.value) {
      // (X == C) => (X == C) and (X != C) => (X != C).
      if ((val.compareTo(equation.val) == 0) && (equal == equation.equal)) {
        return true;
      }
      // (X == C1) => (X != C2), where C1 != C2.
      if ((val.compareTo(equation.val) != 0) && equal && !equation.equal) {
        return true;
      }
    }

    if (!value && !equation.value) {
      final V equationRhs = lhs.equals(equation.lhs) ? equation.rhs : equation.lhs;

      // (X == Y) => (X == Y) and (X != Y) => (X != Y).
      if (rhs.equals(equationRhs) && (equal == equation.equal)) {
        return true;
      }
    }

    return false;
  }

  @SuppressWarnings("unchecked")
  @Override
  public boolean equals(final Object o) {
    if (o == this) {
      return true;
    }

    if (o == null || !(o instanceof IntegerEquation)) {
      return false;
    }

    final IntegerEquation<V> r = (IntegerEquation<V>) o;

    return equal == r.equal && value == r.value && lhs.equals(r.lhs) &&
        (!value && rhs.equals(r.rhs) || value && val.equals(r.val));
  }

  @Override
  public int hashCode() {
    return (31 * lhs.hashCode() + (value ? val.hashCode() : rhs.hashCode()) << 1) ^
        (equal ? 0 : -1);
  }

  @Override
  public String toString() {
    return String.format("%s %s %s", lhs, (equal ? "==" : "!="),
        (value ? val.toString(16) : rhs.toString()));
  }
}
